@@ -31,10 +31,19 @@
 ## -- 2021-12-31  1.7.4     DA       Class Log: udpated docstrings
 ## -- 2022-07-27  1.7.5     DA       A little refactoring
 ## -- 2022-08-21  1.7.6     DA       A little refactoring
+## -- 2022-10-02  1.8.0     DA       Class Log:
+## --                                - new methods Log.get_name(), Log.set_name()
+## --                                - method log(): C_NAME in quotation marks
+## -- 2022-10-29  1.8.1     DA       Class Log: removed call of switch_logging() from __init__()
+## -- 2022-11-04  1.8.2     DA       Class Timer: refactoring
+## -- 2022-11-07  1.9.0     DA       Class Log: new method get_log_level()
+## -- 2023-01-14  1.9.1     SY       Add class Label
+## -- 2023-01-31  1.9.2     SY       Renaming class Label to PersonalisedStamp
+## -- 2023-02-22  2.0.0     DA       Class Saveable: new custom method _save()
 ## -------------------------------------------------------------------------------------------------
 
 """
-Ver. 1.7.6 (2022-08-21)
+Ver. 2.0.0 (2023-02-22)
 
 This module provides various classes with elementry functionalities for reuse in higher level classes. 
 For example: logging, load/save, timer...
@@ -45,6 +54,7 @@ from datetime import datetime, timedelta
 from time import sleep
 import dill as pkl
 import os
+import uuid
 from mlpro.bf.exceptions import *
 
 
@@ -71,8 +81,7 @@ class Loadable:
             A loaded object, if file content was loaded successfully. None otherwise.
         """
 
-        if not os.path.exists(p_path + os.sep + p_filename):
-            return None
+        if not os.path.exists(p_path + os.sep + p_filename): return None
 
         return pkl.load(open(p_path + os.sep + p_filename, 'rb'))
 
@@ -95,7 +104,8 @@ class Saveable:
         """
         To be redefined in case of use of internal generated file names.
 
-        Returns: 
+        Returns
+        -------
             Returns an internal unique filename. 
         """
 
@@ -106,13 +116,17 @@ class Saveable:
     def save(self, p_path, p_filename=None) -> bool:
         """
         Saves content to the given path and file name. If file name is None, a unique file name will
-        be generated by calling method generate_filename(). If it returns False then the saving method is failed. 
+        be generated by calling method generate_filename(). If it returns False then the saving method 
+        is failed. For saving the object, the custom method _save is called. This contains a default
+        implementation based on pickle/dill and can be redefined on demand.
 
-        Parameters:
+        Parameters
+        ----------
             p_path          Path where file will be saved
             p_filename      File name (if None an internal filename will be generated)
 
-        Returns: 
+        Returns
+        -------
             True, if file content was saved successfully. False otherwise.
         """
 
@@ -124,13 +138,23 @@ class Saveable:
         if self.filename is None:
             return False
 
-        try:
-            if not os.path.exists(p_path):
-                os.makedirs(p_path)
-            pkl.dump(self, open(p_path + os.sep + self.filename, "wb"))
-            return True
-        except:
-            return False
+        if not os.path.exists(p_path):
+            os.makedirs(p_path)
+
+        return self._save( p_path=p_path, p_filename=self.filename)
+
+
+## -------------------------------------------------------------------------------------------------
+    def _save(self, p_path, p_filename) -> bool:
+        """
+        Custom method for saving an object. The default implementation is based on pickle/dill. Redefine
+        on demand. See method save() for further details.
+        """
+
+        pkl.dump( obj=self, 
+                  file=open(p_path + os.sep + p_filename, "wb"),
+                  protocol=pkl.HIGHEST_PROTOCOL )
+        return True
 
 
 
@@ -193,12 +217,25 @@ class Log:
 
 ## -------------------------------------------------------------------------------------------------
     def __init__(self, p_logging=C_LOG_ALL):
-        self.switch_logging(p_logging)
+        if p_logging not in self.C_LOG_LEVELS: 
+            raise ParamError('Wrong log level. See class Log for valid log levels')
+        self._level = p_logging
+
         if self.C_INST_MSG:
             self.log(self.C_LOG_TYPE_I, 'Instantiated')
             self.C_INST_MSG = False
 
 
+## -------------------------------------------------------------------------------------------------
+    def get_name(self) -> str:
+        return self.C_NAME
+
+
+## -------------------------------------------------------------------------------------------------
+    def set_name(self, p_name:str):
+        self.C_NAME = p_name
+
+
 ## -------------------------------------------------------------------------------------------------
     def switch_logging(self, p_logging):
         """
@@ -214,6 +251,11 @@ class Log:
         if p_logging not in self.C_LOG_LEVELS: raise ParamError('Wrong log level. See class Log for valid log levels')
         self._level = p_logging
 
+
+ ## -------------------------------------------------------------------------------------------------
+    def get_log_level(self):
+        return self._level
+
  
  ## -------------------------------------------------------------------------------------------------
     def log(self, p_type, *p_args):
@@ -249,7 +291,7 @@ class Log:
 
         print(col + '%04d-%02d-%02d  %02d:%02d:%02d.%06d ' % (
         now.year, now.month, now.day, now.hour, now.minute, now.second, now.microsecond),
-              p_type + '  ' + self.C_TYPE + ' ' + self.C_NAME + ':', *p_args, self.C_COL_RESET)
+              p_type + '  ' + self.C_TYPE + ' "' + self.C_NAME + '":', *p_args, self.C_COL_RESET)
 
 
 
@@ -260,30 +302,32 @@ class Log:
 class Timer:
     """
     Timer class in two time modes (real/virtual) and with simple lap management.
+
+    Parameters
+    ----------
+    p_mode : int         
+        C_MODE_REAL for real time mode or C_MODE_VIRTUAL for virtual time mode
+    p_lap_duration : timedelta = None
+        Optional duration of a single lap.
+    p_lap_limit : int = C_LAP_LIMIT    
+        Maximum number of laps until the lap counter restarts with 0  
     """
 
-    C_MODE_REAL = 0  # Real time
-    C_MODE_VIRTUAL = 1  # Virtual time
+    C_MODE_REAL         = 0  # Real time
+    C_MODE_VIRTUAL      = 1  # Virtual time
 
-    C_LAP_LIMIT = 999999
+    C_LAP_LIMIT         = 999999
 
 ## -------------------------------------------------------------------------------------------------
-    def __init__(self, p_mode, p_lap_duration: timedelta, p_lap_limit=C_LAP_LIMIT) -> None:
-        """
-        Parameters:
-            p_mode          C_MODE_REAL for real time mode,
-                            C_MODE_VIRTUAL for virtual time mode
-            p_lap_duration  Duration of a single lap
-            p_lap_limit     maximum number of laps until the lap counter restarts with 0  
-        """
+    def __init__(self, p_mode:int, p_lap_duration:timedelta=None, p_lap_limit:int=C_LAP_LIMIT):
 
-        self.mode = p_mode
-        self.lap_duration = p_lap_duration
+        self._mode = p_mode
+        self._lap_duration = p_lap_duration
 
         if p_lap_limit == 0:
-            self.lap_limit = self.C_LAP_LIMIT
+            self._lap_limit = self.C_LAP_LIMIT
         else:
-            self.lap_limit = p_lap_limit
+            self._lap_limit = p_lap_limit
 
         self.reset()
 
@@ -301,7 +345,7 @@ class Timer:
         self.lap_time = timedelta(0, 0, 0)
         self.lap_id = 0
 
-        if self.mode == self.C_MODE_REAL:
+        if self._mode == self.C_MODE_REAL:
             self.timer_start_real = datetime.now()
             self.lap_start_real = self.timer_start_real
             self.time_real = self.timer_start_real
@@ -309,7 +353,7 @@ class Timer:
 
 ## -------------------------------------------------------------------------------------------------
     def get_time(self) -> timedelta:
-        if self.mode == self.C_MODE_REAL:
+        if self._mode == self.C_MODE_REAL:
             self.time_real = datetime.now()
             self.time = self.time_real - self.timer_start_real
 
@@ -318,7 +362,7 @@ class Timer:
 
 ## -------------------------------------------------------------------------------------------------
     def get_lap_time(self) -> timedelta:
-        if self.mode == self.C_MODE_REAL:
+        if self._mode == self.C_MODE_REAL:
             self.lap_time = datetime.now() - self.lap_start_real
 
         return self.lap_time
@@ -331,7 +375,7 @@ class Timer:
 
 ## -------------------------------------------------------------------------------------------------
     def add_time(self, p_delta: timedelta):
-        if self.mode == self.C_MODE_VIRTUAL:
+        if self._mode == self.C_MODE_VIRTUAL:
             self.lap_time = self.lap_time + p_delta
             self.time = self.time + p_delta
 
@@ -350,23 +394,24 @@ class Timer:
         timeout = False
 
         # Compute delay until next lap
-        delay = self.lap_duration - self.get_lap_time()
+        if self._lap_duration is not None:
+            delay = self._lap_duration - self.get_lap_time()
 
-        # Check for timeout
-        if delay < timedelta(0, 0, 0):
-            timeout = True
-            delay = timedelta(0, 0, 0)
+            # Check for timeout
+            if delay < timedelta(0, 0, 0):
+                timeout = True
+                delay = timedelta(0, 0, 0)
 
-        # Handle delay depending on timer mode
-        if self.mode == self.C_MODE_REAL:
-            # Wait until next lap start
-            sleep(delay.total_seconds())
-        else:
-            # Just set next lap start time
-            self.time = self.time + delay
+            # Handle delay depending on timer mode
+            if self._mode == self.C_MODE_REAL:
+                # Wait until next lap start
+                sleep(delay.total_seconds())
+            else:
+                # Just set next lap start time
+                self.time = self.time + delay
 
         # Update lap data
-        self.lap_id = divmod(self.lap_id + 1, self.lap_limit)[1]
+        self.lap_id = divmod(self.lap_id + 1, self._lap_limit)[1]
         self.lap_time = timedelta(0, 0, 0)
         self.lap_start_real = datetime.now()
 
@@ -441,4 +486,101 @@ class  ScientificObject:
     C_SCIREF_INSTITUTION = None
     C_SCIREF_CONFERENCE = None
     C_SCIREF_NOTES = None
-    C_SCIREF_EDITOR = None
\ No newline at end of file
+    C_SCIREF_EDITOR = None
+
+
+
+
+## -------------------------------------------------------------------------------------------------
+## -------------------------------------------------------------------------------------------------
+
+class PersonalisedStamp:
+    """
+    This class serves as a base class of label to set up a name and id for another class.
+    
+    Parameters
+    ----------
+    p_name : str
+        name of the created class.
+    p_id : int
+        unique id of the created class. Default: None.
+        
+    Attributes
+    ----------
+    C_NAME : str
+        name of the created class. Default: ''.
+    """
+
+    C_NAME = ''
+
+## -------------------------------------------------------------------------------------------------
+    def __init__(self, p_name:str, p_id:int=None):
+
+        self.C_NAME = p_name
+
+        if p_name != '':
+            self.set_name(p_name)
+        else:
+            raise NotImplementedError('Please add a name!')
+        
+        self.set_id(p_id)
+            
+
+## -------------------------------------------------------------------------------------------------
+    def set_id(self, p_id:int=None):
+        """
+        This method provides a functionality to set an unique ID.
+
+        Parameters
+        ----------
+        p_id : int, optional
+            An unique ID. Default: None.
+
+        """
+        if p_id is None:
+            self._id = str(uuid.uuid4())
+        else:
+            self._id = str(p_id)
+
+
+## -------------------------------------------------------------------------------------------------
+    def get_id(self) -> str:
+        """
+        This method provides a functionality to get the defined unique ID.
+
+        Returns
+        -------
+        str
+            The unique ID.
+
+        """
+        return self._id
+
+
+## -------------------------------------------------------------------------------------------------
+    def set_name(self, p_name:str):
+        """
+        This method provides a functionality to set an unique name.
+
+        Parameters
+        ----------
+        p_name : str
+            An unique name.
+
+        """
+        self._name = p_name
+        self.C_NAME = p_name
+
+
+## -------------------------------------------------------------------------------------------------
+    def get_name(self) -> str:
+        """
+        This method provides a functionality to get the unique name.
+
+        Returns
+        -------
+        str
+            The unique name of the related component.
+
+        """
+        return self._name
\ No newline at end of file
