@@ -38,43 +38,59 @@
 ## --                                  'Score(MA) until Stagnation'
 ## -- 2022-02-27  1.7.1     SY       Refactoring due to auto generated ID in class Dimension
 ## -- 2022-05-23  1.7.2     SY       Bug fixing: storing data reward
+## -- 2022-11-01  1.7.3     DA       Refactoring and code cleaning
+## -- 2022-11-02  1.8.0     DA       Refactoring: methods adapt(), _adapt()
+## -- 2022-11-07  1.8.1     DA       Class RLScenario:
+## --                                - method _run_cycle(): new return value "end_of_data"
+## --                                - method _setup(): refactoring
+## -- 2022-11-09  1.8.2     DA       Refactoring and code cleaning
+## -- 2022-11-29  1.8.3     DA       Refactoring
+## -- 2023-02-02  1.8.4     DA       Class RLTraining: signature of method init_plot refactored
+## -- 2023-02-20  1.9.0     DA       Class RLScenario: new methods load(), _save()
+## -- 2023-03-09  1.9.1     DA       Class RLTrainingResults: removed parameter p_path
 ## -------------------------------------------------------------------------------------------------
 
 """
-Ver. 1.7.2 (2022-05-23)
+Ver. 1.9.1 (2023-03-09)
 
 This module provides model classes to define and run rl scenarios and to train agents inside them.
 """
 
+
+import dill as pkl
+from mlpro.bf.data import DataStoring
+from mlpro.bf.math import *
+from mlpro.bf.ml import *
 from mlpro.rl.models_env import *
 
 
 
+
 ## -------------------------------------------------------------------------------------------------
 ## -------------------------------------------------------------------------------------------------
-class RLDataStoring(DataStoring):
+class RLDataStoring (DataStoring):
     """
     Derivative of basic class DataStoring that is specialized to store episodic training data in the
     context of reinforcement learning.
+
+    Parameters
+    ----------
+    p_space : Set         
+        Space object that provides dimensional information for raw data. If None, a training header 
+        data object will be instantiated.
     """
 
     # Frame ID renamed
-    C_VAR0 = 'Episode ID'
+    C_VAR0          = 'Episode ID'
 
     # Variables for episodic detail data storage
-    C_VAR_CYCLE = 'Cycle'
-    C_VAR_DAY = 'Day'
-    C_VAR_SEC = 'Second'
-    C_VAR_MICROSEC = 'Microsecond'
+    C_VAR_CYCLE     = 'Cycle'
+    C_VAR_DAY       = 'Day'
+    C_VAR_SEC       = 'Second'
+    C_VAR_MICROSEC  = 'Microsecond'
 
-    ## -------------------------------------------------------------------------------------------------
+ ## -------------------------------------------------------------------------------------------------
     def __init__(self, p_space: Set = None):
-        """
-        Parameters:
-            p_space         Space object that provides dimensional information for raw data. If None
-                            a training header data object will be instantiated.
-        """
-
         self.space = p_space
 
         # Initialization as an episodic detail data storage
@@ -89,20 +105,24 @@ class RLDataStoring(DataStoring):
 
         super().__init__(self.variables)
 
-    ## -------------------------------------------------------------------------------------------------
+
+## -------------------------------------------------------------------------------------------------
     def get_variables(self):
         return self.variables
 
-    ## -------------------------------------------------------------------------------------------------
+
+## -------------------------------------------------------------------------------------------------
     def get_space(self):
         return self.space
 
-    ## -------------------------------------------------------------------------------------------------
+
+## -------------------------------------------------------------------------------------------------
     def add_episode(self, p_episode_id):
         self.add_frame(p_episode_id)
         self.current_episode = p_episode_id
 
-    ## -------------------------------------------------------------------------------------------------
+
+## -------------------------------------------------------------------------------------------------
     def memorize_row(self, p_cycle_id, p_tstamp: timedelta, p_data):
         """
         Memorizes an episodic data row.
@@ -122,6 +142,9 @@ class RLDataStoring(DataStoring):
             self.memorize(var, self.current_episode, p_data[i])
 
 
+
+
+
 ## -------------------------------------------------------------------------------------------------
 ## -------------------------------------------------------------------------------------------------
 class RLDataStoringEval(DataStoring):
@@ -134,7 +157,6 @@ class RLDataStoringEval(DataStoring):
     p_space : Set
         Set object that provides dimensional information for raw data. If None a training header data object will
     be instantiated.
-
     """
 
     # Frame ID renamed
@@ -151,7 +173,7 @@ class RLDataStoringEval(DataStoring):
     C_VAR_NUM_LIMIT             = 'Timeouts'                    # Number of timeouts (cycles per episode reached)
     C_VAR_NUM_ADAPT             = 'Adaptations'                 # Number of adaptations in the last training period
 
-    ## -------------------------------------------------------------------------------------------------
+ ## -------------------------------------------------------------------------------------------------
     def __init__(self, p_space: Set):
 
         self.space = p_space
@@ -176,20 +198,24 @@ class RLDataStoringEval(DataStoring):
 
         super().__init__(self.variables)
 
-    ## -------------------------------------------------------------------------------------------------
+
+## -------------------------------------------------------------------------------------------------
     def get_variables(self):
         return self.variables
 
-    ## -------------------------------------------------------------------------------------------------
+
+## -------------------------------------------------------------------------------------------------
     def get_space(self):
         return self.space
 
-    ## -------------------------------------------------------------------------------------------------
+
+## -------------------------------------------------------------------------------------------------
     def add_evaluation(self, p_evaluation_id):
         self.add_frame(p_evaluation_id)
         self.current_evaluation = p_evaluation_id
 
-    ## -------------------------------------------------------------------------------------------------
+
+## -------------------------------------------------------------------------------------------------
     def memorize_row(self, 
                      p_score, 
                      p_score_ma, 
@@ -252,33 +278,64 @@ class RLDataStoringEval(DataStoring):
             self.memorize(var, self.current_evaluation, p_reward[i])
 
 
+
+
+
 ## -------------------------------------------------------------------------------------------------
 ## -------------------------------------------------------------------------------------------------
-class RLScenario(Scenario):
+class RLScenario (Scenario):
     """
     Template class for an RL scenario consisting of an environment and an agent. 
+
+    Parameters
+    ----------
+    p_mode
+        Operation mode. See bf.ops.Mode.C_VALID_MODES for valid values. Default = Mode.C_MODE_SIM.
+    p_ada : bool
+        Boolean switch for adaptivitiy. Default = True.
+    p_cycle_limit : int
+        Maximum number of cycles (0=no limit, -1=get from env). Default = 0.
+    p_visualize : bool
+        Boolean switch for env/agent visualisation. Default = False.
+    p_logging
+        Log level (see constants of class mlpro.bf.various.Log). Default = Log.C_LOG_WE.
+
+    Attributes
+    ----------
+    C_TYPE : str
+        Constant class type for logging: 'RL-Scenario'.
+    C_NAME : str
+        Constant custom name for logging. To be set in own child class.
     """
 
     C_TYPE = 'RL-Scenario'
     C_NAME = '????'
 
-    ## -------------------------------------------------------------------------------------------------
+ ## -------------------------------------------------------------------------------------------------
     def __init__(self,
-                 p_mode=Mode.C_MODE_SIM,  # Operation mode (see class Mode)
-                 p_ada: bool = True,  # Boolean switch for adaptivity of internal model
-                 p_cycle_limit=0,  # Maximum number of cycles (0=no limit, -1=get from env)
-                 p_visualize=True,  # Boolean switch for env/agent visualisation
-                 p_logging=Log.C_LOG_ALL):  # Log level (see constants of class Log)
+                 p_mode=Mode.C_MODE_SIM,  
+                 p_ada:bool=True,  
+                 p_cycle_limit=0, 
+                 p_visualize:bool=True,  
+                 p_logging=Log.C_LOG_ALL):  
 
         # 1 Setup entire scenario
-        self._env = None
-        super().__init__(p_mode=p_mode, p_ada=p_ada, p_cycle_limit=p_cycle_limit, p_visualize=p_visualize,
-                         p_logging=p_logging)
+        self._env : Environment = None
+
+        super().__init__( p_mode=p_mode, 
+                          p_ada=p_ada, 
+                          p_cycle_limit=p_cycle_limit, 
+                          p_visualize=p_visualize,
+                          p_logging=p_logging )
+
         if self._env is None:
             raise ImplementationError('Please bind your RL environment to self._env')
 
         self._agent = self._model
 
+        self._model_class = self._model.__class__ 
+        self._env_class   = self._env.__class__
+
         # 2 Finalize cycle limit
         if self._cycle_limit == -1:
             self._cycle_limit = self._env.get_cycle_limit()
@@ -286,48 +343,106 @@ class RLScenario(Scenario):
         # 3 Init data logging
         self.connect_data_logger()
 
-    ## -------------------------------------------------------------------------------------------------
+
+## -------------------------------------------------------------------------------------------------
+    @staticmethod
+    def load(p_path):
+        scenario = pkl.load(open(p_path + os.sep + 'scenario', 'rb'))
+        scenario._model = scenario._model_class.load(p_path, 'model')
+        scenario._agent = scenario._model
+        scenario._env   = scenario._env_class.load(p_path, 'environment')
+        return scenario
+
+
+## -------------------------------------------------------------------------------------------------
+    def _save(self, p_path, p_filename=None) -> bool:
+
+        # 1 Create new subfolder for scenario files
+        scenario_path = p_path + os.sep + 'scenario'
+        os.makedirs(scenario_path)
+
+        # 2 Save model and environment
+        self._model.save(scenario_path, 'model')
+        self._env.save(scenario_path, 'environment')
+
+        # 3 Save scenario w/o model and env
+        self._model = None 
+        self._agent = None
+        self._env   = None
+        pkl.dump( obj=self, 
+                  file=open(scenario_path + os.sep + 'scenario', "wb"),
+                  protocol=pkl.HIGHEST_PROTOCOL )
+
+        # 4 Reload model and env
+        self._model = self._model_class.load(scenario_path, 'model')
+        self._agent = self._model
+        self._env   = self._env_class.load(scenario_path, 'environment')
+
+        return True
+
+
+## -------------------------------------------------------------------------------------------------
     def switch_logging(self, p_logging):
         super().switch_logging(p_logging)
         self._env.switch_logging(p_logging)
 
-    ## -------------------------------------------------------------------------------------------------
-    def _setup(self, p_mode, p_ada: bool, p_logging: bool) -> Model:
+
+## -------------------------------------------------------------------------------------------------
+    def _setup(self, p_mode, p_ada: bool, p_visualize:bool, p_logging) -> Model:
         """
-        Set up the ML scenario by redefinition. Please bind your environment to self._env and return
+        Custom method to set up the ML scenario. Please bind your environment to self._env and return
         the agent as model. 
 
-        Parameters:
-            p_mode          Operation mode (see class Mode)
-            p_ada           Boolean switch for adaptivity of internal model
-            p_logging       Boolean switch for logging functionality
+        Parameters
+        ----------
+        p_mode
+            Operation mode. See Mode.C_VALID_MODES for valid values. Default = Mode.C_MODE_SIM
+        p_ada : bool
+            Boolean switch for adaptivity.
+        p_visualize : bool
+            Boolean switch for env/agent visualisation. Default = True.
+        p_logging
+            Log level (see constants of class Log). 
 
-        Returns:
+        Returns
+        -------
+        agent : Agent
             Agent model (object of type Agent or Multi-agent)
         """
 
         raise NotImplementedError
 
-    ## -------------------------------------------------------------------------------------------------
-    def init_plot(self, p_figure=None):
-        super().init_plot(p_figure=p_figure)
-        self._env.init_plot(p_figure=p_figure)
 
-    ## -------------------------------------------------------------------------------------------------
-    def update_plot(self):
-        super().update_plot()
-        self._env.update_plot()
+## -------------------------------------------------------------------------------------------------
+    def init_plot( self, 
+                   p_figure: Figure = None, 
+                   p_plot_settings: PlotSettings = None ):
 
-    ## -------------------------------------------------------------------------------------------------
+        super().init_plot( p_figure = p_figure, 
+                           p_plot_settings = p_plot_settings )
+
+        self._env.init_plot( p_figure = p_figure, 
+                             p_plot_settings = p_plot_settings )
+
+
+## -------------------------------------------------------------------------------------------------
+    def update_plot(self, **p_kwargs):
+        super().update_plot( **p_kwargs)
+        self._env.update_plot( **p_kwargs)
+
+
+## -------------------------------------------------------------------------------------------------
     def _set_mode(self, p_mode):
         if isinstance(self._env, Mode):
             self._env.set_mode(p_mode)
 
-    ## -------------------------------------------------------------------------------------------------
+
+## -------------------------------------------------------------------------------------------------
     def get_latency(self) -> timedelta:
         return self._env.get_latency()
 
-    ## -------------------------------------------------------------------------------------------------
+
+## -------------------------------------------------------------------------------------------------
     def _reset(self, p_seed):
         """
         Environment and timer are reset. The random generators for environment and agent will
@@ -343,22 +458,26 @@ class RLScenario(Scenario):
         if self._visualize:
             self._env.init_plot()
 
-    ## -------------------------------------------------------------------------------------------------
+
+## -------------------------------------------------------------------------------------------------
     def get_agent(self):
         return self._agent
 
-    ## -------------------------------------------------------------------------------------------------
+
+## -------------------------------------------------------------------------------------------------
     def get_env(self):
         return self._env
 
-    ## -------------------------------------------------------------------------------------------------
+
+## -------------------------------------------------------------------------------------------------
     def connect_data_logger(self, p_ds_states: RLDataStoring = None, p_ds_actions: RLDataStoring = None,
                             p_ds_rewards: RLDataStoring = None):
         self._ds_states = p_ds_states
         self._ds_actions = p_ds_actions
         self._ds_rewards = p_ds_rewards
 
-    ## -------------------------------------------------------------------------------------------------
+
+## -------------------------------------------------------------------------------------------------
     def _run_cycle(self):
         """
         Processes a single cycle.
@@ -371,9 +490,14 @@ class RLScenario(Scenario):
             True on error. False otherwise.
         adapted : bool
             True, if agent adapted something in this cycle. False otherwise.
-
+        end_of_data : bool
+            True, if the end of the related data source has been reached. False otherwise.
         """
 
+        # 0 Initialization
+        end_of_data = False
+
+
         # 1 Environment: get current state
         state = self._env.get_state()
         state.set_tstamp(self._timer.get_time())
@@ -381,6 +505,7 @@ class RLScenario(Scenario):
         if self._ds_states is not None:
             self._ds_states.memorize_row(self._cycle_id, self._timer.get_time(), state.get_values())
 
+
         # 2 Agent: compute and log next action
         self.log(self.C_LOG_TYPE_I, 'Process time', self._timer.get_time(), ': Agent computes action...')
         action = self._agent.compute_action(state)
@@ -389,12 +514,14 @@ class RLScenario(Scenario):
         if self._ds_actions is not None:
             self._ds_actions.memorize_row(self._cycle_id, ts, action.get_sorted_values())
 
+
         # 3 Environment: process agent's action
         self.log(self.C_LOG_TYPE_I, 'Process time', self._timer.get_time(), ': Env processes action...')
         self._env.process_action(action)
         self._timer.add_time(self._env.get_latency())  # in virtual mode only...
         self._env.get_state().set_tstamp(self._timer.get_time())
 
+
         # 4 Environment: compute and log reward
         reward = self._env.compute_reward()
         ts = self._timer.get_time()
@@ -408,9 +535,11 @@ class RLScenario(Scenario):
 
                 self._ds_rewards.memorize_row(self._cycle_id, ts, reward_values)
 
+
         # 5 Agent: adapt policy
         self.log(self.C_LOG_TYPE_I, 'Process time', self._timer.get_time(), ': Agent adapts policy...')
-        adapted = self._agent.adapt(self._env.get_state(), reward)
+        adapted = self._agent.adapt(p_state=self._env.get_state(), p_reward=reward)
+
 
         # 6 Check for terminating events
         success = self._env.get_state().get_success()
@@ -422,7 +551,10 @@ class RLScenario(Scenario):
         if error:
             self.log(self.C_LOG_TYPE_E, 'Process time', self._timer.get_time(), ': Environment terminated')
 
-        return success, error, adapted
+        return success, error, adapted, end_of_data
+
+
+
 
 
 ## -------------------------------------------------------------------------------------------------
@@ -439,8 +571,6 @@ class RLTrainingResults(TrainingResults):
         Run id.
     p_cycle_id : int
         Id of first cycle of this run.
-    p_path : str
-        Optional destination path to store the results.
     p_logging
         Log level (see constants of class Log). Default: Log.C_LOG_ALL
 
@@ -456,9 +586,12 @@ class RLTrainingResults(TrainingResults):
     C_CPAR_NUM_EPI = 'Training Episodes'
     C_CPAR_NUM_EVAL = 'Evaluations'
 
-    ## -------------------------------------------------------------------------------------------------
-    def __init__(self, p_scenario: RLScenario, p_run, p_cycle_id, p_path=None, p_logging=Log.C_LOG_WE):
-        super().__init__(p_scenario, p_run, p_cycle_id, p_path=p_path, p_logging=p_logging)
+## -------------------------------------------------------------------------------------------------
+    def __init__(self, p_scenario: RLScenario, p_run, p_cycle_id, p_logging=Log.C_LOG_WE):
+        super().__init__( p_scenario = p_scenario, 
+                          p_run = p_run, 
+                          p_cycle_id = p_cycle_id, 
+                          p_logging = p_logging )
 
         self.num_episodes = 0
         self.num_evaluations = 0
@@ -467,20 +600,23 @@ class RLTrainingResults(TrainingResults):
         self.ds_rewards = None
         self.ds_eval = None
 
-    ## -------------------------------------------------------------------------------------------------
+
+## -------------------------------------------------------------------------------------------------
     def close(self):
         super().close()
 
         self.add_custom_result(self.C_CPAR_NUM_EPI, self.num_episodes)
         self.add_custom_result(self.C_CPAR_NUM_EVAL, self.num_evaluations)
 
-    ## -------------------------------------------------------------------------------------------------
+
+## -------------------------------------------------------------------------------------------------
     def _log_results(self):
         super()._log_results()
         self.log(self.C_LOG_TYPE_W, '-- Training Episodes :', self.num_episodes)
         self.log(self.C_LOG_TYPE_W, '-- Evaluations       :', self.num_evaluations)
 
-    ## -------------------------------------------------------------------------------------------------
+
+## -------------------------------------------------------------------------------------------------
     def save(self, p_path, p_filename='summary.csv') -> bool:
         if not super().save(p_path, p_filename=p_filename):
             return False
@@ -495,9 +631,12 @@ class RLTrainingResults(TrainingResults):
             self.ds_eval.save_data(p_path, self.C_FNAME_EVAL)
 
 
+
+
+
 ## -------------------------------------------------------------------------------------------------
 ## -------------------------------------------------------------------------------------------------
-class RLTraining(Training):
+class RLTraining (Training):
     """
     This class performs an episodic training on a (multi-)agent in a given environment. Both are
     expected as parts of a reinforcement learning scenario (see class RLScenario for more details).
@@ -545,14 +684,13 @@ class RLTraining(Training):
         Boolean switch for env/agent visualisation. Default = False.
     p_logging
         Log level (see constants of class mlpro.bf.various.Log). Default = Log.C_LOG_WE.
-
     """
 
     C_NAME = 'RL'
 
     C_CLS_RESULTS = RLTrainingResults
 
-    ## -------------------------------------------------------------------------------------------------
+## -------------------------------------------------------------------------------------------------
     def __init__(self, **p_kwargs):
 
         # 1 Initialization of elementary training functionalities
@@ -683,7 +821,8 @@ class RLTraining(Training):
                 self._counter_epi_eval = 0
                 self._mode = self.C_MODE_EVAL
 
-    ## -------------------------------------------------------------------------------------------------
+
+## -------------------------------------------------------------------------------------------------
     def _init_results(self) -> TrainingResults:
         results = super()._init_results()
 
@@ -717,7 +856,8 @@ class RLTraining(Training):
 
         return results
 
-    ## -------------------------------------------------------------------------------------------------
+
+## -------------------------------------------------------------------------------------------------
     def _init_episode(self):
 
         # 1 Evaluation handling  
@@ -769,7 +909,8 @@ class RLTraining(Training):
         if (self._results.ds_rewards and self._scenario._ds_rewards) is not None:
             self._results.ds_rewards.add_episode(self._results.num_episodes)
 
-    ## -------------------------------------------------------------------------------------------------
+
+## -------------------------------------------------------------------------------------------------
     def _close_episode(self):
 
         if self._eval_frequency > 0:
@@ -820,7 +961,8 @@ class RLTraining(Training):
 
         self._cycles_episode = 0
 
-    ## -------------------------------------------------------------------------------------------------
+
+## -------------------------------------------------------------------------------------------------
     def _init_evaluation(self):
         """
         Initializes the next evaluation.
@@ -838,7 +980,8 @@ class RLTraining(Training):
         self._eval_num_broken = 0
         self._eval_sum_reward = None
 
-    ## -------------------------------------------------------------------------------------------------
+
+## -------------------------------------------------------------------------------------------------
     def _update_evaluation(self, p_success: bool, p_error: bool, p_cycle_limit: bool):
         """
         Updates evaluation statistics.
@@ -890,7 +1033,8 @@ class RLTraining(Training):
         else:
             raise Error('Reward type ' + str(reward_type) + ' not yet supported')
 
-    ## -------------------------------------------------------------------------------------------------
+
+## -------------------------------------------------------------------------------------------------
     def _close_evaluation(self) -> np.ndarray:
         """
         Closes the current evaluation and computes a related score.
@@ -957,7 +1101,8 @@ class RLTraining(Training):
         # 6 Outro
         return score
 
-    ## -------------------------------------------------------------------------------------------------
+
+## -------------------------------------------------------------------------------------------------
     def _run_cycle(self) -> bool:
         """
         Runs single training cycle.
@@ -975,7 +1120,7 @@ class RLTraining(Training):
             self._init_episode()
 
         # 2 Run a cycle
-        success, error, timeout, limit, adapted = self._scenario.run_cycle()
+        success, error, timeout, limit, adapted, end_of_data = self._scenario.run_cycle()
         self._cycles_episode += 1
 
         if adapted:
