@@ -16,10 +16,11 @@
 ## -- 2022-03-24  1.0.6     SY       Refactoring
 ## -- 2022-03-25  1.0.7     SY       Change methods names (SetupSpaces to setup_spaces)
 ## -- 2022-08-14  1.1.0     DA       Introduction of root class Wrapper
+## -- 2022-10-17  1.1.1     SY       Refactoring due to unit test
 ## -------------------------------------------------------------------------------------------------
 
 """
-Ver. 1.1.0 (2022-08-14)
+Ver. 1.1.1 (2022-10-17)
 
 This module provides a wrapper class for hyperparameter tuning by reusing the Hyperopt framework.
 
@@ -120,17 +121,17 @@ class WrHPTHyperopt(HyperParamTuner, Wrapper, ScientificObject):
         if self._num_trials <= 0:
             raise ParamError('Parameter self._num_trials must be greater than 0')
         
-        if self._root_path is None:
-            raise ParamError('Mandatory parameter self._root_path is not supplied')
-        
         if self._training_param is None:
             raise ParamError('Mandatory parameter self._training_param is not supplied')
         
         # change root path in training param
-        self._training_param['p_training_param']['p_path'] = self._root_path+os.sep+'HyperparameterTuning'+os.sep+'Base (Preparation)'
-        if not os.path.exists(self._training_param['p_training_param']['p_path']):
-            os.mkdir(self._root_path+os.sep+'HyperparameterTuning')
-            os.mkdir(self._training_param['p_training_param']['p_path'])
+        if self._root_path is None:
+            self._training_param['p_training_param']['p_path'] = None
+        else:
+            self._training_param['p_training_param']['p_path'] = self._root_path+os.sep+'HyperparameterTuning'+os.sep+'Base (Preparation)'
+            if not os.path.exists(self._training_param['p_training_param']['p_path']):
+                os.mkdir(self._root_path+os.sep+'HyperparameterTuning')
+                os.mkdir(self._training_param['p_training_param']['p_path'])
         
         # ignore collecting data during tuning to save tuning time and memory
         self._training_param['p_training_param']['p_collect_states'] = False
@@ -161,7 +162,8 @@ class WrHPTHyperopt(HyperParamTuner, Wrapper, ScientificObject):
         trials = Trials()
         best_param = fmin(self._ofct_hyperopt, spaces, self.algo, self._num_trials, trials=trials)
         best_result = trials.results[np.argmin([r['loss'] for r in trials.results])]['loss']
-        self.save(best_param, -best_result, 'best_parameters.csv')
+        if self._root_path is not None:
+            self.save(best_param, -best_result, 'best_parameters.csv')
         
         return -best_result
 
@@ -183,9 +185,12 @@ class WrHPTHyperopt(HyperParamTuner, Wrapper, ScientificObject):
         self.log(self.C_LOG_TYPE_I, self.C_LOG_SEPARATOR, '\n')
         
         # change root path in training param
-        self._training_param['p_training_param']['p_path'] =  self._root_path+os.sep+'HyperparameterTuning'+os.sep+'Trial_'+str(self.num_trials)
-        if not os.path.exists(self._training_param['p_training_param']['p_path']):
-            os.mkdir(self._training_param['p_training_param']['p_path'])
+        if self._root_path is None:
+            self._training_param['p_training_param']['p_path'] = None
+        else:
+            self._training_param['p_training_param']['p_path'] =  self._root_path+os.sep+'HyperparameterTuning'+os.sep+'Trial_'+str(self.num_trials)
+            if not os.path.exists(self._training_param['p_training_param']['p_path']):
+                os.mkdir(self._training_param['p_training_param']['p_path'])
         
         # instantiate a scenario class
         training_cls = self._training_cls(**self._training_param['p_training_param'])
